import importlib
from dataclasses import dataclass
from threading import Thread
from typing import Iterator

import torch
from openai import OpenAI
from tenacity import retry, stop_after_attempt, wait_random_exponential
from transformers import AutoModelForCausalLM, AutoTokenizer, TextIteratorStreamer

from amt_pipeline import LOGGER


@dataclass
class BaseNLP:
    """Base class for language processing"""

    device: torch.device = torch.device("cuda", 0) if torch.cuda.is_available() else torch.device("cpu")
    dtype: torch.dtype = torch.bfloat16 if torch.cuda.is_bf16_supported() else torch.float16

    def generate(self, chat_messages: list[dict[str, str]]) -> Iterator[str]:
        """
        Generate text.

        Args:
            - `chat_messages`: list, formatted list of interactions

        Returns:
            - str, streamed generation output text
        """
        NotImplementedError()

    def apply_prompt_and_generate(
        self,
        input_text: str,
        task_name: str,
        output_lang: str,
        chat_messages: list[dict[str, str]] = [],
    ) -> Iterator[str]:
        """
        Apply a prompt related to the task and generate related text.

        Args:
            - `input_text`: str, the document or user input
            - `task_name`: str, the name of the task to apply
            - `output_lang`: str, the output text language
            - `chat_messages`: list, previous conversation messages (if any)

        Returns:
            - str, streamed generation output text
        """
        supported_tasks = list(self.prompts.keys())
        if task_name not in supported_tasks:
            raise UserWarning(f"Supported tasks for {self.__class__.__name__} are {supported_tasks}")

        supported_lang = list(self.prompts[task_name].keys())
        if output_lang not in supported_lang:
            LOGGER.warn(
                f"[{self.__class__.__name__}] {task_name} only supports {supported_lang} languages. Defaulting to en."
            )
            output_lang = "en"

        text_l = len("".join([msg["content"] for msg in chat_messages] + [input_text])) * 1e-3
        LOGGER.info(f"[{self.__class__.__name__}] Received {text_l:.2f}k text. Generate {task_name} in {output_lang}")

        system_prompt = self.prompts[task_name][output_lang]
        chat_messages = (
            [{"role": "system", "content": system_prompt}] + chat_messages + [{"role": "user", "content": input_text}]
        )

        return self.generate(chat_messages)


class ShisaNLP(BaseNLP):
    """Japanese/English LLM based on [shisa](https://github.com/AUGMXNT/shisa)"""

    prompts: dict[str, dict] = {
        "assistant": {
            "en": "You are a highly skilled AI trained in language comprehension and assistance. I would like you to read the following text and assist the user's query regarding this text. Aim to answer as helpfully as possible and if you don't know the answer to a question, please don't share false information.",
            "ja": "あなたは、日本語の理解と支援を訓練された高度な日本人AIです。 以下のテキストを読んで、このテキストに関するユーザーの問い合わせを手伝ってほしい。 できるだけ役に立つように答え、質問の答えが分からない場合は、間違った情報を共有しないでください。",
        },
        "translation": {
            "en": "You are a highly skilled AI trained in language comprehension and translation. I would like you to read the following text and translate it in English. Aim to keep a similar structure as the original text so that the meaning of the text stays the same.",
            "ja": "あなたは言語理解と翻訳の訓練を受けた高度なAIです。 次の文章を読んで日本語で翻訳してほしいです。 原文と同様の構造を保ち、テキストの意味が同じになるようにすることを目指します。",
        },
        "summary": {
            "en": "You are a highly skilled AI trained in language comprehension and summarization. I would like you to read the following text and summarize it into a concise abstract paragraph. Aim to retain the most important points, providing a coherent and readable summary that could help a person understand the main points of the discussion without needing to read the entire text. Please avoid unnecessary details or tangential points.",
            "ja": "あなたは、日本語の理解と要約の訓練を受けた高度な日本人AIです。 次の文章を読んで、簡潔な抽象的な段落に要約してほしい。 最も重要な点を保持し、全体のテキストを読むことなく、人が議論の要点を理解するのに役立つ一貫した読みやすい要約を提供することを目的とします。 不必要な詳細や接線は避けてください。",
        },
        "keywords-3": {
            "en": "You are a highly skilled AI trained in language comprehension and key points extraction. I would like you to read the following text and extract 3 main key points about it. Aim to only retain the most important points and avoid unnecessary details or tangential points.",
            "ja": "あなたは、日本語の理解力と要点抽出を訓練された高度な日本人AIです。 以下の文章を読んで、それに関する3つの主要なポイントを抽出してほしい。 最も重要な点のみを保持し、不必要な詳細や接線点を避けることを目指します。",
        },
    }

    def __init__(self, model_name: str = "augmxnt/shisa-7b-v1") -> None:
        use_flash_attn = True

        if importlib.util.find_spec("flash_attn") is None:
            use_flash_attn = False
            LOGGER.warn(f"You can accelerate inference for {self.__class__.__name__} with the `flash_attn` package")

        self.tokenizer = AutoTokenizer.from_pretrained(model_name, use_fast=True)
        self.model = AutoModelForCausalLM.from_pretrained(
            model_name,
            torch_dtype=self.dtype,
            low_cpu_mem_usage=True,
            device_map={"": self.device},
            attn_implementation="flash_attention_2" if use_flash_attn else None,
        ).eval()
        self.streamer = TextIteratorStreamer(self.tokenizer, skip_prompt=True, skip_special_tokens=True)

    @torch.no_grad
    def generate(self, chat_messages: list[dict[str, str]]) -> Iterator[str]:
        """
        Generate text.

        Args:
            - `chat_messages`: list, formatted list of interactions

        Returns:
            - str, streamed generation output text
        """
        inputs = self.tokenizer.apply_chat_template(chat_messages, add_generation_prompt=True, return_tensors="pt")

        generation_kwargs = dict(
            inputs=inputs.to(self.model.device),
            pad_token_id=self.tokenizer.eos_token_id,
            max_new_tokens=1024,
            do_sample=True,
            temperature=0.5,
            repetition_penalty=1.15,
            top_p=0.95,
            streamer=self.streamer,
        )
        thread = Thread(target=self.model.generate, kwargs=generation_kwargs)
        thread.start()

        for stream in self.streamer:
            yield stream


class OpenaiNLP(BaseNLP):
    """Models using the OpenAI API"""

    prompts: dict[str, dict] = {
        "assistant": {
            "en": "You are a highly skilled AI trained in language comprehension and assistance. I would like you to read the following text and assist the user's query regarding this text. Aim to answer as helpfully as possible and if you don't know the answer to a question, please don't share false information.",
            "ja": "You are a highly skilled Japanese AI trained in Japanese language comprehension and assistance. I would like you to read the following text and assist the user's query regarding this text. Aim to answer as helpfully as possible and if you don't know the answer to a question, please don't share false information.",
        },
        "translation": {
            "en": "You are a highly skilled AI trained in language comprehension and translation. I would like you to read the following text and translate it in English. Aim to keep a similar structure as the original text so that the meaning of the text stays the same.",
            "ja": "You are a highly skilled AI trained in language comprehension and translation. I would like you to read the following text and translate it in Japanese. Aim to keep a similar structure as the original text so that the meaning of the text stays the same.",
        },
        "summary": {
            "en": "You are a highly skilled AI trained in language comprehension and summarization. I would like you to read the following text and summarize it into a concise abstract paragraph. Aim to retain the most important points, providing a coherent and readable summary that could help a person understand the main points of the discussion without needing to read the entire text. Please avoid unnecessary details or tangential points.",
            "ja": "You are a highly skilled Japanese AI trained in Japanese language comprehension and summarization. I would like you to read the following text and summarize it into a concise abstract paragraph. Aim to retain the most important points, providing a coherent and readable summary that could help a person understand the main points of the discussion without needing to read the entire text. Please avoid unnecessary details or tangential points.",
        },
        "keywords-3": {
            "en": "You are a highly skilled AI trained in language comprehension and key points extraction. I would like you to read the following text and extract 3 main key points about it. Aim to only retain the most important points and avoid unnecessary details or tangential points.",
            "ja": "You are a highly skilled Japanese AI trained in Japanese language comprehension and key points extraction. I would like you to read the following text and extract 3 main key points about it. Aim to only retain the most important points and avoid unnecessary details or tangential points.",
        },
    }

    def __init__(self, model_name: str = "gpt-3.5-turbo-0125"):
        self.model_name = model_name
        self.client = OpenAI()

    @retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
    def completion_with_backoff(self, *args, **kwargs):
        return self.client.chat.completions.create(*args, **kwargs)

    def generate(self, chat_messages: list[dict[str, str]]) -> Iterator[str]:
        """
        Generate text.

        Args:
            - `chat_messages`: list, formatted list of interactions

        Returns:
            - str, streamed generation output text
        """
        stream = self.completion_with_backoff(
            model=self.model_name,
            messages=chat_messages,
            stream=True,
        )

        for chunk in stream:
            if chunk.choices[0].delta.content is not None:
                yield chunk.choices[0].delta.content
