from dataclasses import dataclass
from typing import Iterator

import torch
from pyannote.core import Segment

from amt_pipeline import AUDIO
from amt_pipeline.outputs import DiarizationResult, PipelineResult
from amt_pipeline.speaker_diarization.diarization import BaseDiarization
from amt_pipeline.speech_recognition.asr import BaseASR


@dataclass
class BasePipeline:
    """Base class for pipelines"""

    def apply(self, speech: torch.Tensor, rate: int, *args, **kwargs) -> Iterator[PipelineResult]:
        """
        Apply the pipeline

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio

        Returns:
            - PipeResult, pipeline result
        """
        NotImplementedError()


class DiarizeTranscribePipeline(BasePipeline):
    """Transcribe multi-speakers speech"""

    def __init__(self, diarization_model: BaseDiarization, asr_model: BaseASR) -> None:
        self.diarization = diarization_model
        self.asr = asr_model

    def apply(  # noqa: C901
        self,
        speech: torch.Tensor,
        rate: int,
        min_segment_duration: float = 0.0,
        max_segment_interval: float = 0.0,
        condition_on_previous: bool = False,
        diarization_kwargs: dict = {},
        asr_kwargs: dict = {},
    ) -> Iterator[PipelineResult]:
        """
        Apply the pipeline

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio
            - `min_segment_duration`: float, the min number of seconds for a detected segment (shorter are dropped)
            - `min_segment_interval`: float, the max number of seconds interval between two consecutive segments of the same speaker to merge
            - `condition_on_previous`: bool, if True, conditions the ASR on the previous utterance
            - `diarization_kwargs`: dict, optional arguments for the diarization model
            - `asr_kwargs`: dict, optional arguments for the asr model

        Returns:
            - PipeResult, pipeline result
        """
        diarization_result = self.diarization.diarize(speech, rate, **diarization_kwargs)
        audio_duration = AUDIO.get_duration(file={"waveform": speech, "sample_rate": rate})

        previous_iter = [None] + diarization_result[:-1]
        current_iter = diarization_result
        next_iter = diarization_result[1:] + [None]

        previous_holder = None  # replace previous if there is a merge between the previous and current
        current_holder = None  # replace current if there is a merge between the current and next

        previous_result = None  # for prompting the ASR model

        for previous, current, next in zip(previous_iter, current_iter, next_iter):
            if previous_holder is not None:
                previous = previous_holder
            if current_holder is not None:
                current = current_holder

            if previous is not None:
                # merge previous and current
                if ((current.start - previous.end) < max_segment_interval) and (
                    previous.speaker_id == current.speaker_id
                ):
                    current = DiarizationResult(start=previous.start, end=current.end, speaker_id=current.speaker_id)
                    previous_holder = current
                else:
                    previous_holder = None

            if next is not None:
                # merge current and next, skip the iteration
                if ((next.start - current.end) < max_segment_interval) and (current.speaker_id == next.speaker_id):
                    current = DiarizationResult(start=current.start, end=next.end, speaker_id=current.speaker_id)
                    current_holder = current
                    previous_holder = previous
                    continue
                else:
                    current_holder = None

            segment = Segment(current.start, min(current.end, audio_duration))
            segment_speech, segment_rate = AUDIO.crop(file={"waveform": speech, "sample_rate": rate}, segment=segment)

            if (segment.duration < min_segment_duration) or (segment_speech.shape[-1] < 1):
                continue

            if condition_on_previous and (previous_result is not None):
                if previous_result.speaker_id == current.speaker_id:
                    asr_kwargs["conditional_prompt"] = previous_result.text
                else:
                    asr_kwargs["conditional_prompt"] = None

            asr_result = self.asr.transcribe(segment_speech, segment_rate, **asr_kwargs)

            result = PipelineResult(
                text=asr_result.text,
                lang=asr_result.lang,
                start=current.start,
                end=current.end,
                speaker_id=current.speaker_id,
            )

            previous_result = result

            yield result
