from dataclasses import dataclass

import numpy as np
import pandas as pd
import torch
from pyannote.audio import Inference, Model, Pipeline
from pyannote.core import Annotation, SlidingWindow, SlidingWindowFeature

from amt_pipeline import AUDIO, LOGGER, validate_kwargs
from amt_pipeline.outputs import DiarizationResult


@dataclass
class BaseDiarization:
    """Base class for Diarization models"""

    device: torch.device = torch.device("cuda", 0) if torch.cuda.is_available() else torch.device("cpu")

    def diarize(self, speech: torch.Tensor, rate: int, *args, **kwargs) -> list[DiarizationResult]:
        """
        Segments the audio and attribute a speaker to each of them.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio

        Returns:
            - DiarizationResult, Diarization result
        """
        NotImplementedError()


class PyannoteDiarization(BaseDiarization):
    """Diarization models from the [pyannote.audio](https://github.com/pyannote/pyannote-audio) toolkit"""

    def __init__(self, model_name: str = "pyannote/speaker-diarization-3.1") -> None:
        self.model = Pipeline.from_pretrained(model_name).to(self.device)

    @validate_kwargs
    def diarize(self, speech: torch.Tensor, rate: int, num_speakers: int = None) -> list[DiarizationResult]:
        """
        Segments the audio and attribute a speaker to each of them.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio
            - `num_speakers`: int, number of speakers in the audio speech

        Returns:
            - list, list of Diarization result
        """
        if not num_speakers:
            num_speakers = None
            LOGGER.info(f"[{self.__class__.__name__}] No num speakers received. Default to {num_speakers}")

        audio_s = speech.shape[-1] / rate
        LOGGER.info(f"[{self.__class__.__name__}] Received {audio_s:.2f}s of audio. Diarize {num_speakers} speakers")

        output = self.model({"waveform": speech, "sample_rate": rate}, num_speakers=num_speakers)
        converted_output = self.__annotation_to_pandas(output)

        result = [DiarizationResult(**row) for row in converted_output.to_dict(orient="records")]
        return result

    def __annotation_to_pandas(self, annotation: Annotation) -> pd.DataFrame:
        """
        Convert `pyannote` Annotation object to a `pandas` DataFrame object.

        Args:
            - `annotation`: Annotation, the annotation to convert

        Returns:
            - DataFrame, the converted dataframe
        """
        df = (
            pd.DataFrame.from_records([x.split(" ") for x in annotation.to_rttm().split("\n")])
            .rename(columns={3: "start", 4: "duration", 7: "speaker_id"})
            .dropna()
        )
        df[["start", "duration"]] = df[["start", "duration"]].astype(float)
        df["end"] = df["start"] + df["duration"]
        df = df.drop(columns=[0, 1, 2, 5, 6, 8, 9, "duration"])

        return df


class LocalGlobalDiarization(BaseDiarization):
    """Local Global old pipeline"""

    def __init__(self) -> None:
        from amt_pipeline.legacy.clustering import AgglomerativeClustering
        from amt_pipeline.segmentation import PowersetSegmentation

        self.seg_model = PowersetSegmentation(device=self.device)
        self.cluster_model = AgglomerativeClustering(device=self.device)

    @validate_kwargs
    def diarize(self, speech: torch.Tensor, rate: int, num_speakers: int = None) -> list[DiarizationResult]:
        """
        Segments the audio and attribute a speaker to each of them.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio
            - `num_speakers`: int, number of speakers in the audio speech

        Returns:
            - list, list of Diarization result
        """
        if not num_speakers:
            num_speakers = None
            LOGGER.info(f"No num speakers received for {self.__class__.__name__}. Default to {num_speakers}")

        LOGGER.info(f"{self.__class__.__name__}: diarize speech with {num_speakers} speakers")

        inputs_dict = {"waveform": speech, "sample_rate": rate}
        active_segments = []

        # segmentation (i.e., local diarization)
        for segments in self.seg_model.segment(**inputs_dict, min_seconds=0.1):
            active_segments.extend([segment for segment in segments if len(segment["local_speakers"]) > 0])

        # filter segments for embedding model
        # embedding model needs at least 0.11 seconds of audio or will return NaNs
        filtered_segments = [x for x in active_segments if x["end"] - x["start"] >= 0.11]

        segment_wav = [segment["waveform"] for segment in filtered_segments]
        segment_sr = [segment["sample_rate"] for segment in filtered_segments]

        # clustering (i.e., global diarization
        speakers = self.cluster_model.predict(segment_wav, segment_sr, num_speakers=num_speakers)
        diarized_segments = [{**segment, **dict(speaker=spk)} for segment, spk in zip(filtered_segments, speakers)]
        result = [DiarizationResult(start=x["start"], end=x["end"], speaker_id=x["speaker"]) for x in diarized_segments]

        return result


class SegmentClusterDiarization(BaseDiarization):
    """Diarization model that uses segmentation (local diarization) and clustering (global diarization)"""

    def __init__(
        self,
        segmentation_model_name: str = "pyannote/segmentation-3.0",
        embedding_model_name: str = "pyannote/wespeaker-voxceleb-resnet34-LM",
        clustering_model_name: str = "",
    ) -> None:
        self.segmentation_model = Inference(
            segmentation_model_name,
            window="sliding",
            skip_aggregation=True,
            skip_conversion=False,
            batch_size=8,
            device=self.device,
        )
        self.embedding_model = Inference(
            Model.from_pretrained(embedding_model_name), window="whole", device=self.device
        )

        self.window_duration = self.segmentation_model.model.specifications.duration
        # self.clustering_model = sklearn.cluster.AgglomerativeClustering(n_clusters=3)

    def diarize(self, speech: torch.Tensor, rate: int) -> DiarizationResult:
        """
        Segments the audio and attribute a speaker to each of them.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio

        Returns:
            - list, list of Diarization result
        """
        segmentation_results = self.segmentation_model({"waveform": speech, "sample_rate": rate})
        input(segmentation_results.data.shape)

        local_speaker_list: list = []

        for window_segment, features in segmentation_results:
            frames_per_window, n_classes = features.shape
            frame_duration = window_segment.duration / frames_per_window
            frames = SlidingWindow(step=frame_duration, duration=frame_duration, end=window_segment.duration)

            for local_speaker in range(n_classes):
                speaker_segments = []
                for i, frame in enumerate(frames):
                    if features[i, local_speaker] > 0:
                        window, sr = AUDIO.crop({"waveform": speech, "sample_rate": rate}, frame, mode="pad")
                        speaker_segments.append(window)
                if len(speaker_segments) > 0:
                    speaker_segments = torch.nested.nested_tensor(speaker_segments)
                    speaker_segments = torch.nested.to_padded_tensor(speaker_segments, 0)
                    print(speaker_segments.shape)
                    local_speaker_list.append(speaker_segments)

        input(len(local_speaker_list))
        local_speaker_list = torch.nested.nested_tensor(local_speaker_list)
        local_speaker_list = torch.nested.to_padded_tensor(local_speaker_list, 0)
        print(local_speaker_list.shape)
        # # (start, end, n_speakers) of each frame
        # count = SlidingWindowFeature(data=np.sum(features, axis=-1, keepdims=True), sliding_window=new_frame)
        # input(count.data.shape)

        # for segment in frames:
        #     new_window, sr = AUDIO.crop(
        #         {"waveform": window, "sample_rate": sr}, segment, duration=segment.duration, mode="pad"
        #     )
        #     print(segment, features.shape, new_window.shape)
        #     input()

        input()

        n_window, frames_per_window, n_classes = segmentation_results.data.shape
        frame_duration = self.window_duration / frames_per_window
        frames = SlidingWindow(step=frame_duration, duration=frame_duration)  # (start, end) times of each frame

        count = self.speaker_count(segmentation_results, frames=frames)  # (start, end, n_speakers) of each frame
        input(count.data.shape)
        local_diarization_results = self.to_diarization(
            segmentation_results, count
        )  # (start, end, local_diarization) of each frame
        input(local_diarization_results.data.shape)

        # speech_to_embed = []
        # for segment, feature in local_diarization_results:
        #     if segment.start < duration:
        #         print(speech.shape, duration, segment.start, segment.end)
        #         wav, sr = AUDIO.crop({"waveform": speech, "sample_rate": rate}, segment, duration=frame_duration, mode="pad")
        #         print(wav.shape, sr, segment)
        #         speech_to_embed.append(wav)
        # speech_to_embed = [t.reshape(1, -1) for t in torch.cat(speech_to_embed).split(2)]
        # for x in speech_to_embed:
        #     print(x.shape)

        # embeddings = [self.embedding_model({"waveform": wav, "sample_rate": self.model_rate}) for wav in speech_to_embed]
        # embeddings = np.stack(embeddings)
        # print(embeddings.shape)

    @staticmethod
    def speaker_count(
        binarized_segmentations: SlidingWindowFeature,
        frames: SlidingWindow = None,
    ) -> SlidingWindowFeature:
        """Estimate frame-level number of instantaneous speakers

        Parameters
        ----------
        binarized_segmentations : SlidingWindowFeature
            (num_chunks, num_frames, num_classes)-shaped binarized scores.
        frames : SlidingWindow, optional
            Frames resolution. Defaults to estimate it automatically based on
            `segmentations` shape and chunk size. Providing the exact frame
            resolution (when known) leads to better temporal precision.

        Returns
        -------
        count : SlidingWindowFeature
            (num_frames, 1)-shaped instantaneous speaker count
        """

        count = Inference.aggregate(
            np.sum(binarized_segmentations, axis=-1, keepdims=True),
            frames=frames,
            hamming=False,
            missing=0.0,
            skip_average=False,
        )
        count.data = np.rint(count.data).astype(np.uint8)

        return count

    @staticmethod
    def to_diarization(
        segmentations: SlidingWindowFeature,
        count: SlidingWindowFeature,
    ) -> SlidingWindowFeature:
        """Build diarization out of preprocessed segmentation and precomputed speaker count

        Parameters
        ----------
        segmentations : SlidingWindowFeature
            (num_chunks, num_frames, num_speakers)-shaped segmentations
        count : SlidingWindow_feature
            (num_frames, 1)-shaped speaker count

        Returns
        -------
        discrete_diarization : SlidingWindowFeature
            Discrete (0s and 1s) diarization.
        """

        # TODO: investigate alternative aggregation
        activations = Inference.aggregate(
            segmentations,
            frames=count.sliding_window,
            hamming=False,
            missing=0.0,
            skip_average=True,
        )
        # shape is (num_frames, num_speakers)

        _, num_speakers = activations.data.shape
        max_speakers_per_frame = np.max(count.data)
        if num_speakers < max_speakers_per_frame:
            activations.data = np.pad(activations.data, ((0, 0), (0, max_speakers_per_frame - num_speakers)))

        extent = activations.extent & count.extent
        activations = activations.crop(extent, return_data=False)
        count = count.crop(extent, return_data=False)

        sorted_speakers = np.argsort(-activations, axis=-1)
        binary = np.zeros_like(activations.data)

        for t, ((_, c), speakers) in enumerate(zip(count, sorted_speakers)):
            for i in range(c.item()):
                binary[t, speakers[i]] = 1.0

        return SlidingWindowFeature(binary, activations.sliding_window)
