import importlib
from dataclasses import dataclass

import torch
from espnet2.bin.s2t_inference import Speech2Text
from transformers import pipeline

from amt_pipeline import LOGGER, validate_kwargs
from amt_pipeline.outputs import ASRResult


@dataclass
class BaseASR:
    """Base class for ASR models"""

    device: torch.device = torch.device("cuda", 0) if torch.cuda.is_available() else torch.device("cpu")
    dtype: torch.dtype = torch.bfloat16 if torch.cuda.is_bf16_supported() else torch.float16

    def transcribe(self, speech: torch.Tensor, rate: int, *args, **kwargs) -> ASRResult:
        """
        Transcribe the audio speech.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio

        Returns:
            - ASRResult, ASR result
        """
        NotImplementedError()


class EspnetASR(BaseASR):
    """ASR models from the [ESPnet](https://github.com/espnet/espnet) toolkit"""

    lang_to_code: dict[str, str] = {
        "en": "eng",
        "ja": "jpn",
        "zh": "zho",
        "fr": "fra",
    }

    def __init__(self, model_name: str = "espnet/owsm_v3.1_ebf") -> None:
        # TODO: Espnet seems to only handle float32 dtype
        self.dtype = torch.float32
        self.model = Speech2Text.from_pretrained(
            model_tag=model_name,
            device=str(self.device),
            dtype=str(self.dtype).split(".")[-1],
            beam_size=5,
            ctc_weight=0.0,
            maxlenratio=0.0,
        )

    @validate_kwargs
    def transcribe(
        self,
        speech: torch.Tensor,
        rate: int,
        input_lang: str = "en",
        output_lang: str = "en",
        conditional_prompt: str = None,
    ) -> ASRResult:
        """
        Transcribe the audio speech.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio
            - `input_lang`: str, language of the audio speech
            - `output_lang`: str, language of the transcribed text
            - `conditional_prompt`: str, conditional text prompt for generation

        Returns:
            - ASRResult, ASR result
        """
        if not input_lang:
            input_lang = "en" if not output_lang else output_lang
            LOGGER.info(f"[{self.__class__.__name__}] No input language received. Default to {input_lang}")

        if not output_lang:
            output_lang = "en" if not input_lang else input_lang
            LOGGER.info(f"[{self.__class__.__name__}] No output language received. Default to {output_lang}")

        supported_lang = list(self.lang_to_code.keys())
        if (input_lang not in supported_lang) or (output_lang not in supported_lang):
            raise UserWarning(f"Supported languages for {self.__class__.__name__} are {supported_lang}")

        audio_s = speech.shape[-1] / rate
        LOGGER.info(
            f"[{self.__class__.__name__}] Received {audio_s:.2f}s of audio. Transcribe {input_lang} -> {output_lang}"
        )

        if input_lang == output_lang:
            task_sym = "<asr>"
        else:
            task_sym = f"<st_{self.lang_to_code[output_lang]}>"

        model_max_length = int(self.model.preprocessor_conf["fs"] * self.model.preprocessor_conf["speech_length"])
        chunks = speech.split(model_max_length, dim=-1)
        texts = [conditional_prompt]

        for chunk in chunks:
            outputs = self.model(
                chunk.squeeze(),
                lang_sym=f"<{self.lang_to_code[input_lang]}>",
                task_sym=task_sym,
                predict_time=False,
                text_prev=texts[-1],
            )
            text = outputs[0][-2].strip()
            texts.append(text)

        output_text = " ".join(texts[1:])

        return ASRResult(text=output_text, lang=output_lang)


class WhisperASR(BaseASR):
    """Whisper ASR model loaded from [transformers](https://github.com/huggingface/transformers)"""

    lang_to_code: dict[str, str] = {
        "en": "english",
        "ja": "japanese",
        "zh": "chinese",
        "fr": "french",
    }

    def __init__(self, model_name: str = "openai/whisper-large-v3") -> None:
        use_flash_attn = True

        if importlib.util.find_spec("flash_attn") is None:
            use_flash_attn = False
            LOGGER.warn(f"You can accelerate inference for {self.__class__.__name__} with the `flash_attn` package")

        self.model = pipeline(
            "automatic-speech-recognition",
            model=model_name,
            torch_dtype=self.dtype,
            chunk_length_s=30,
            model_kwargs={
                "low_cpu_mem_usage": True,
                "device_map": {"": self.device},
                "attn_implementation": "flash_attention_2" if use_flash_attn else None,
            },
        )

    @validate_kwargs
    def transcribe(
        self,
        speech: torch.Tensor,
        rate: int,
        input_lang: str = "en",
        output_lang: str = "en",
        conditional_prompt: str = None,
    ) -> ASRResult:
        """
        Transcribe the audio speech.

        Args:
            - `speech`: Tensor, the audio speech
            - `rate`: int, the sample rate of the audio
            - `input_lang`: str, language of the audio speech
            - `output_lang`: str, language of the transcribed text
            - `conditional_prompt`: str, conditional text prompt for generation

        Returns:
            - ASRResult, ASR result
        """
        if not input_lang:
            input_lang = "en" if not output_lang else output_lang
            LOGGER.info(f"[{self.__class__.__name__}] No input language received. Default to {input_lang}")

        if not output_lang:
            output_lang = "en" if not input_lang else input_lang
            LOGGER.info(f"[{self.__class__.__name__}] No output language received. Default to {output_lang}")

        supported_lang = list(self.lang_to_code.keys())
        if (input_lang not in supported_lang) or (output_lang not in supported_lang):
            raise UserWarning(f"Supported languages for {self.__class__.__name__} are {supported_lang}")

        audio_s = speech.shape[-1] / rate
        LOGGER.info(
            f"[{self.__class__.__name__}] Received {audio_s:.2f}s of audio. Transcribe {input_lang} -> {output_lang}"
        )

        if (input_lang != output_lang) and (output_lang == "en"):  # translate only to en
            task = "translate"
        else:
            task = "transcribe"

        generate_kwargs = {
            "task": task,
            "language": output_lang,
            "condition_on_prev_tokens": None,  # TODO: check if True change something (warnings say not)
            "prompt_condition_type": "first-segment",
        }

        if conditional_prompt is not None:
            prompt_ids = self.model.tokenizer.get_prompt_ids(conditional_prompt, return_tensors="pt")
            prompt_ids = prompt_ids[-self.model.model.config.max_target_positions :]
            generate_kwargs["prompt_ids"] = prompt_ids.to(self.model.device)

        outputs = self.model(
            {"raw": speech.squeeze().numpy(), "sampling_rate": rate},
            generate_kwargs=generate_kwargs,
        )

        text = outputs["text"].strip()

        if conditional_prompt is not None:
            text = text[len(conditional_prompt) :]  # noqa: E203

        return ASRResult(text=text, lang=output_lang)
